#ifndef MMTRACE
#include <stdio.h>
#include <string.h>
#endif
#include "shader_common.h"
#include "nv30_shader.h"

unsigned int NVFP_TX_AOP_COUNT = 64;
struct _op_xlat NVFP_TX_AOP[64];

static struct _op_xlat *
NV30FPGetOPTXRec(ShaderFunc *shader, int merged)
{
	int op;

	op = shader->GetOpcodeHW(shader, 0);
	if (op > NVFP_TX_AOP_COUNT)
		return NULL;
	if (NVFP_TX_AOP[op].SOP == SOP_UNKNOWN)
		return NULL;
	return &NVFP_TX_AOP[op];
}

static int
NV30FPHasMergedInst(ShaderFunc *shader)
{
	return 0;
}

static int
NV30FPIsLastInst(ShaderFunc *shader)
{
	return ((shader->inst[0] & NV30_FP_OP_PROGRAM_END) ? 1 : 0);
}

static int
NV30FPGetOffsetNext(ShaderFunc *shader)
{
	int i;

	for (i=0;i<3;i++)
		if (shader->GetSourceFile(shader, 0, i) == SRF_CONST)
			return 8;

	return 4;
}

static ShaderOpcode
NV30FPGetOpcode(ShaderFunc *shader, int merged)
{
	struct _op_xlat *opr;

	opr = shader->GetOPTXRec(shader, merged);
	if (!opr)
		return SOP_UNKNOWN;

	return opr->SOP;
}

static unsigned int
NV30FPGetOpcodeHW(ShaderFunc *shader, int slot)
{
	int op;

	op = (shader->inst[0] & NV30_FP_OP_OPCODE_MASK)	>> NV30_FP_OP_OPCODE_SHIFT;

	return op;
}

static ShaderRegFile
NV30FPGetDestFile(ShaderFunc *shader, int merged)
{
	/* Result regs overlap temporary regs */
	return SRF_TEMP;
}

static unsigned int
NV30FPGetDestID(ShaderFunc *shader, int merged)
{
	int id;

	switch (shader->GetDestFile(shader, merged)) {
	case SRF_TEMP:
		id = ((shader->inst[0] & NV30_FP_OP_OUT_REG_MASK)
				>> NV30_FP_OP_OUT_REG_SHIFT);
		return id;
	default:
		return -1;
	}
}

static unsigned int
NV30FPGetDestMask(ShaderFunc *shader, int merged)
{
	unsigned int mask = 0;

	if (shader->inst[0] & NV30_FP_OP_OUT_X) mask |= SMASK_X;
	if (shader->inst[0] & NV30_FP_OP_OUT_Y) mask |= SMASK_Y;
	if (shader->inst[0] & NV30_FP_OP_OUT_Z) mask |= SMASK_Z;
	if (shader->inst[0] & NV30_FP_OP_OUT_W) mask |= SMASK_W;

	return mask;
}

static unsigned int
NV30FPGetSourceHW(ShaderFunc *shader, int merged, int pos)
{
	struct _op_xlat *opr;

	opr = shader->GetOPTXRec(shader, merged);
	if (!opr || opr->srcpos[pos] == -1)
		return -1;

	return shader->inst[opr->srcpos[pos] + 1];
}

static ShaderRegFile
NV30FPGetSourceFile(ShaderFunc *shader, int merged, int pos)
{
	unsigned int src;
	struct _op_xlat *opr;
	int file;

	opr = shader->GetOPTXRec(shader, merged);
	if (!opr || opr->srcpos[pos] == -1)
		return SRF_UNKNOWN;

	switch (opr->srcpos[pos]) {
	case SPOS_TEXUNIT: return SRF_TEXUNIT;
	case SPOS_ADDRESS: return SRF_ADDRESS;
	default:
		src  = shader->GetSourceHW(shader, merged, pos);
		file = (src & NV30_FP_REG_TYPE_MASK) >> NV30_FP_REG_TYPE_SHIFT;

		switch (file) {
		case NV30_FP_REG_TYPE_TEMP : return SRF_TEMP;
		case NV30_FP_REG_TYPE_INPUT: return SRF_ATTRIB;
		case NV30_FP_REG_TYPE_CONST: return SRF_CONST;
		default:
			return SRF_UNKNOWN;
		}
	}
}

static int
NV30FPGetSourceID(ShaderFunc *shader, int merged, int pos)
{
	switch (shader->GetSourceFile(shader, merged, pos)) {
	case SRF_ATTRIB:
		switch ((shader->inst[0] & NV30_FP_OP_INPUT_SRC_MASK)
				>> NV30_FP_OP_INPUT_SRC_SHIFT) {
		case NV30_FP_OP_INPUT_SRC_POSITION: return SR_POSITION;
		case NV30_FP_OP_INPUT_SRC_COL0  : return SR_COL0;
		case NV30_FP_OP_INPUT_SRC_COL1  : return SR_COL1;
		case NV30_FP_OP_INPUT_SRC_FOGC  : return SR_FOGCOORD;
		case NV30_FP_OP_INPUT_SRC_TC(0) : return SR_TEXCOORD0;
		case NV30_FP_OP_INPUT_SRC_TC(1) : return SR_TEXCOORD1;
		case NV30_FP_OP_INPUT_SRC_TC(2) : return SR_TEXCOORD2;
		case NV30_FP_OP_INPUT_SRC_TC(3) : return SR_TEXCOORD3;
		case NV30_FP_OP_INPUT_SRC_TC(4) : return SR_TEXCOORD4;
		case NV30_FP_OP_INPUT_SRC_TC(5) : return SR_TEXCOORD5;
		case NV30_FP_OP_INPUT_SRC_TC(6) : return SR_TEXCOORD6;
		case NV30_FP_OP_INPUT_SRC_TC(7) : return SR_TEXCOORD7;
		default: return -1;
		}
		break;
	case SRF_TEMP:
	{
		unsigned int src;

		src = shader->GetSourceHW(shader, merged, pos);
		return ((src & NV30_FP_REG_SRC_MASK) >> NV30_FP_REG_SRC_SHIFT);
	}
	case SRF_TEXUNIT:
		return ((shader->inst[0] & NV30_FP_OP_TEX_UNIT_MASK)
				>> NV30_FP_OP_TEX_UNIT_SHIFT);
		break;
	case SRF_CONST: /* inlined into fragprog */
	default:
		return -1;
	}
}

static int
NV30FPGetSourceNegate(ShaderFunc *shader, int merged, int pos)
{
	unsigned int src;

	src = shader->GetSourceHW(shader, merged, pos);

	if (src == -1)
		return -1;
	return ((src & NV30_FP_REG_NEGATE) ? 1 : 0);
}

static int
NV30FPGetSourceAbs(ShaderFunc *shader, int merged, int pos)
{
	struct _op_xlat *opr;
	static unsigned int abspos[3] = {
		NV30_FP_OP_OUT_ABS,
		(1<<30), /* guess */
		(1<<31)  /* guess */
	};

	opr = shader->GetOPTXRec(shader, merged);
	if (!opr || opr->srcpos[pos] == -1)
		return -1;

	return ((shader->inst[1] & abspos[opr->srcpos[pos]]) ? 1 : 0);
}

ShaderSwzComp NV30FP_TX_SWIZZLE[4] = { SSC_X, SSC_Y, SSC_Z, SSC_W };
static ShaderRegSwz
NV30FPTXSwizzle(int hwswz)
{
	ShaderRegSwz swz;

	swz.w = NV30FP_TX_SWIZZLE[(hwswz & 0xC0) >> 6];
	swz.z = NV30FP_TX_SWIZZLE[(hwswz & 0x30) >> 4];
	swz.y = NV30FP_TX_SWIZZLE[(hwswz & 0x0C) >> 2];
	swz.x = NV30FP_TX_SWIZZLE[(hwswz & 0x03) >> 0];

	return swz;
}

static ShaderRegSwz
NV30FPGetSourceSwizzle(ShaderFunc *shader, int merged, int pos)
{
	unsigned int src;
	int swz;

	src = shader->GetSourceHW(shader, merged, pos);
	swz = (src & NV30_FP_REG_SWZ_ALL_MASK) >> NV30_FP_REG_SWZ_ALL_SHIFT;
	return NV30FPTXSwizzle(swz);
}

static int
NV30FPGetSourceIndexed(ShaderFunc *shader, int merged, int pos)
{
	switch (shader->GetSourceFile(shader, merged, pos)) {
	case SRF_ATTRIB:
		return ((shader->inst[3] & NV30_FP_OP_INDEX_INPUT) ? 1 : 0);
	default:
		return 0;
	}
}

static void
NV30FPGetSourceConstVal(ShaderFunc *shader, int merged, int pos, float *val)
{
	val[0] = *(float *)&(shader->inst[4]);
	val[1] = *(float *)&(shader->inst[5]);
	val[2] = *(float *)&(shader->inst[6]);
	val[3] = *(float *)&(shader->inst[7]);
}

static int
NV30FPGetSourceScale(ShaderFunc *shader, int merged, int pos)
{
/*FIXME: is this per-source, only for a specific source, or all sources??*/
	return (1 << ((shader->inst[2] & NV30_FP_OP_SRC_SCALE_MASK)
				>> NV30_FP_OP_SRC_SCALE_SHIFT));
}

static int
NV30FPGetAddressRegID(ShaderFunc *shader)
{
	return 0;
}

static ShaderSwzComp
NV30FPGetAddressRegSwizzle(ShaderFunc *shader)
{
	return SSC_X;
}

static int
NV30FPSupportsConditional(ShaderFunc *shader)
{
	/*FIXME: Is this true of all ops? */
	return 1;
}

static int
NV30FPGetConditionUpdate(ShaderFunc *shader)
{
	return ((shader->inst[0] & NV30_FP_OP_COND_WRITE_ENABLE) ? 1 : 0);
}

static int
NV30FPGetConditionTest(ShaderFunc *shader)
{
	/*FIXME: always? */
	return 1;
}

static ShaderCond
NV30FPGetCondition(ShaderFunc *shader)
{
	int cond;

	cond = ((shader->inst[1] & NV30_FP_OP_COND_MASK)
			>> NV30_FP_OP_COND_SHIFT);

	switch (cond) {
	case NV30_FP_OP_COND_FL: return SC_FL;
	case NV30_FP_OP_COND_LT: return SC_LT;
	case NV30_FP_OP_COND_EQ: return SC_EQ;
	case NV30_FP_OP_COND_LE: return SC_LE;
	case NV30_FP_OP_COND_GT: return SC_GT;
	case NV30_FP_OP_COND_NE: return SC_NE;
	case NV30_FP_OP_COND_GE: return SC_GE;
	case NV30_FP_OP_COND_TR: return SC_TR;
	default:
		return SC_UNKNOWN;
	}
}

static ShaderRegSwz
NV30FPGetCondRegSwizzle(ShaderFunc *shader)
{
	int swz;

	swz = (shader->inst[1] & NV30_FP_OP_COND_SWZ_ALL_MASK)
		>> NV30_FP_OP_COND_SWZ_ALL_SHIFT;
	return NV30FPTXSwizzle(swz);
}

static int
NV30FPGetCondRegID(ShaderFunc *shader)
{
	return 0;
}

static ShaderPrecision
NV30FPGetPrecision(ShaderFunc *shader)
{
	int p;

	p = (shader->inst[0] & NV30_FP_OP_PRECISION_MASK)
		>> NV30_FP_OP_PRECISION_SHIFT;

	switch (p) {
		case NV30_FP_PRECISION_FP32: return SP_FLOAT32;
		case NV30_FP_PRECISION_FP16: return SP_FLOAT16;
		case NV30_FP_PRECISION_FX12: return SP_FIXED12;
		default:
			return SP_UNKNOWN;
	}
}

static int
NV30FPGetSaturate(ShaderFunc *shader)
{
	return ((shader->inst[0] & NV30_FP_OP_OUT_SAT) ? 1 : 0);
}

void
NV30FPInitShaderFuncs(ShaderFunc *shader)
{
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_MOV, SOP_MOV, 0, -1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_MUL, SOP_MUL, 0, 1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_ADD, SOP_ADD, 0, 1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_MAD, SOP_MAD, 0, 1, 2 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_DP3, SOP_DP3, 0, 1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_DP4, SOP_DP4, 0, 1, -1 ); 
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_DST, SOP_DST, 0, 1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_MIN, SOP_MIN, 0, 1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_MAX, SOP_MAX, 0, 1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_SLT, SOP_SLT, 0, 1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_SGE, SOP_SGE, 0, 1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_FRC, SOP_FRC, 0, -1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_FLR, SOP_FLR, 0, -1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_TEX, SOP_TEX, 0, 4, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_TXD, SOP_TXD, 0, 1, 2 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_TXP, SOP_TXP, 0, 4, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_TXB, SOP_TXB, 0, 4, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_SEQ, SOP_SEQ, 0, 1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_SGT, SOP_SGT, 0, 1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_SLE, SOP_SLE, 0, 1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_SNE, SOP_SNE, 0, 1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_RCP, SOP_RCP, 0, -1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_LG2, SOP_LG2, 0, -1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_EX2, SOP_EX2, 0, -1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_COS, SOP_COS, 0, -1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_SIN, SOP_SIN, 0, -1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_NOP, SOP_NOP, -1, -1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_DDX, SOP_DDX, 0, -1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_DDY, SOP_DDY, 0, -1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_PK4B, SOP_PK4B, 0, -1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_UP4B, SOP_UP4B, 0, -1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_PK2H, SOP_PK2H, 0, -1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_UP2H, SOP_UP2H, 0, -1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_PK4UB, SOP_PK4UB, 0, -1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_UP4UB, SOP_UP4UB, 0, -1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_PK2US, SOP_PK2US, 0, -1, -1 );
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_UP2US, SOP_UP2US, 0, -1, -1 );
	/*FIXME: Haven't confirmed the source positions for the below opcodes */
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_LIT, SOP_LIT, 0, -1, -1);
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_LRP, SOP_LRP, 0,  1,  2);
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_POW, SOP_POW, 0,  1, -1);
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_RSQ, SOP_RSQ, 0, -1, -1);
	MOD_OPCODE(NVFP_TX_AOP, NV30_FP_OP_OPCODE_RFL, SOP_RFL, 0,  1, -1);


	shader->GetOPTXRec           = NV30FPGetOPTXRec;
	shader->HasMergedInst        = NV30FPHasMergedInst;
	shader->IsLastInst           = NV30FPIsLastInst;
	shader->GetOffsetNext        = NV30FPGetOffsetNext;
	shader->GetOpcode            = NV30FPGetOpcode;
	shader->GetOpcodeHW          = NV30FPGetOpcodeHW;
	shader->GetDestFile          = NV30FPGetDestFile;
	shader->GetDestID            = NV30FPGetDestID;
	shader->GetDestMask          = NV30FPGetDestMask;
	shader->GetSourceHW          = NV30FPGetSourceHW;
	shader->GetSourceFile        = NV30FPGetSourceFile;
	shader->GetSourceID          = NV30FPGetSourceID;
	shader->GetSourceNegate      = NV30FPGetSourceNegate;
	shader->GetSourceAbs         = NV30FPGetSourceAbs;
	shader->GetSourceSwizzle     = NV30FPGetSourceSwizzle;
	shader->GetSourceIndexed     = NV30FPGetSourceIndexed;
	shader->GetSourceConstVal    = NV30FPGetSourceConstVal;
	shader->GetSourceScale       = NV30FPGetSourceScale;
	shader->GetRelAddressRegID   = NV30FPGetAddressRegID;
	shader->GetRelAddressSwizzle = NV30FPGetAddressRegSwizzle;
	shader->GetPrecision         = NV30FPGetPrecision;
	shader->GetSaturate          = NV30FPGetSaturate;
	shader->SupportsConditional  = NV30FPSupportsConditional;
	shader->GetConditionUpdate   = NV30FPGetConditionUpdate;
	shader->GetConditionTest     = NV30FPGetConditionTest;
	shader->GetCondition         = NV30FPGetCondition;
	shader->GetCondRegSwizzle    = NV30FPGetCondRegSwizzle;
	shader->GetCondRegID         = NV30FPGetCondRegID;
}

